(* shared memory (minimalistic version) *)
(* solyga@gmx.de, 2020-05-31 *)
IMPLEMENTATION MODULE ShdMem;

IMPORT WinDef, WinNT, WinBase;
IMPORT Lib, Str, Lib2;
IMPORT Types, SO, Cfg, Log;
FROM SYSTEM IMPORT BYTE, ADDRESS, ADR;
FROM Storage IMPORT ALLOCATE;

CONST
  mid = Cfg.midShdMem;

TYPE
  TResName = ARRAY [0..127] OF CHAR;

TYPE
  THandle = WinDef.HANDLE;

CONST
  hndInv = WinBase.INVALID_HANDLE_VALUE;  (* 0FFFFFFFFH *)

TYPE
  TMutex = THandle;

  TBufDesc = RECORD
    i: TId;
    l: CARDINAL;
    m: TMutex;
    a: ADDRESS;
  END;
  TBuf = POINTER TO TBufDesc;

VAR
  bufs: ARRAY TId OF TBuf;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlMsg, s, FALSE );
END Message;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlFat, s, TRUE );
  HALT;
END Fatal;


PROCEDURE WrHex( VAR b: SO.TBuffer; x-: ARRAY OF BYTE; l: CARDINAL );
CONST
  lMax = 8;
VAR
  i: CARDINAL;
  c: CHAR;
BEGIN
  IF l > SIZE( x ) THEN l := SIZE( x ) END;
  IF l > lMax THEN l := lMax END;
  IF l = 0 THEN RETURN END;
  c := b.c.p;
  b.c.p := '0';
  DEC( l );
  FOR i := 0 TO l DO
    SO.WrShtHex( b, Types.TShtCard( x[i] ), 2 );
    IF i < l THEN SO.WrChar( b, ' ' ) END;
  END;
  b.c.p := c;
END WrHex;


PROCEDURE NewMtx( n-: ARRAY OF CHAR ): TMutex;
VAR
  h: THandle;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  h := WinBase.CreateMutex( NIL, FALSE, n );
  IF h = NIL THEN
    SO.Init( b, s );
    SO.WrStr( b, 'NewMtx: CreateMutex failed, error code ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  RETURN h;
END NewMtx;

PROCEDURE NewShm( n-: ARRAY OF CHAR; l: LONGCARD ): ADDRESS;
VAR
  h: THandle;
  a: ADDRESS;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  h := WinBase.CreateFileMapping(
    hndInv,                (* file handle *)
    NIL,                   (* attributes *)
    WinNT.PAGE_READWRITE,  (* protection *)
    0,                     (* size high *)
    l,                     (* size low, 0 (with high = 0) => WinError.ERROR_INVALID_PARAMETER = 87 *)
    n                      (* name *)
  );
  IF h = NIL THEN
    SO.Init( b, s );
    SO.WrStr( b, 'NewShm: CreateFileMapping failed, error code ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  (* WinError.ERROR_ALREADY_EXISTS = 183 *)
  SO.Init( b, s );
  SO.WrStr( b, 'NewShm: CreateFileMapping succeeded, error code ' );
  SO.WrLngCard( b, WinBase.GetLastError(), 1 );
  Message( s );
  (* docs.microsoft.com: "The initial contents of the pages in a file mapping object
       backed by the operating system paging file are 0 (zero)."
     SOL: It's impossible for the user to safely initialize the memory. OS help is required -
     either with interlocking or with a constant initialization. *)

  a := WinBase.MapViewOfFile(
    h,
    WinBase.FILE_MAP_WRITE,
    0,  (* offset high *)
    0,  (* offset low *)
    l   (* size (still unclear, since 1 works) *)
  );
  IF a = NIL THEN
    SO.Init( b, s );
    SO.WrStr( b, 'NewShm: MapViewOfFile failed, error code ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  SO.Init( b, s );
  SO.WrStr( b, 'NewShm: MapViewOfFile succeeded, adr ' );
  SO.WrLngHex( b, LONGCARD( a ), 8 );
  Message( s );

  RETURN a;
END NewShm;

PROCEDURE MkName( VAR s: ARRAY OF CHAR; n-: ARRAY OF CHAR; x-: ARRAY OF CHAR );
BEGIN
  Str.Copy( s, pfx );
  Str.Append( s, n );
  Str.Append( s, x );
END MkName;

PROCEDURE Open( i: TId ): TBuf;
VAR
  b: TBuf;
  l: CARDINAL;
  n: TResName;
BEGIN
  b := bufs[i];
  IF b # NIL THEN RETURN b END;
  NEW( b );
  Lib2.Fill( b^, 0 );
  b^.i := i;           l := cfg[i].l;
  b^.l := l;           MkName( n, cfg[i].n, sfxMtx );
  b^.m := NewMtx( n ); MkName( n, cfg[i].n, sfxBuf );
  b^.a := NewShm( n, l );
  bufs[i] := b;
  RETURN b;
END Open;


(* the shared memory is not automatically interlocked by win (tested) *)
PROCEDURE Enter( m: TMutex );
VAR
  i: LONGCARD;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  i := WinBase.WaitForSingleObject( m, WinBase.INFINITE );
  IF i = MAX( LONGCARD ) THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Enter: WaitForSingleObject failed, err ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  IF i > 0 THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Enter: WaitForSingleObject returned ' );
    SO.WrLngCard( b, i, 1 );
    Fatal( s );
  END;
END Enter;

PROCEDURE Leave( m: TMutex );
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  IF ~ WinBase.ReleaseMutex( m ) THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Leave: ReleaseMutex failed, err ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
END Leave;


PROCEDURE Write( b: TBuf; x-: ARRAY OF BYTE; l: CARDINAL ): CARDINAL;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Write: ' );
    SO.WrLngCard( b, l, 1 );
    SO.WrStr( b, '  ' );
    WrHex( b, x, l );
    Message( s );
  END Show;

BEGIN
  IF l > SIZE( x ) THEN l := SIZE( x ) END;
  IF b = NIL THEN RETURN l END;
  IF l > b^.l THEN l := b^.l END;
  Enter( b^.m );
  Lib.Move( ADR( x ), b^.a, l );
  Leave( b^.m );
  Show();
  RETURN l;
END Write;

PROCEDURE Read( b: TBuf; VAR x: ARRAY OF BYTE; l: CARDINAL ): CARDINAL;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Read: ' );
    SO.WrLngCard( b, l, 1 );
    SO.WrStr( b, '  ' );
    WrHex( b, x, l );
    Message( s );
  END Show;

BEGIN
  IF l > SIZE( x ) THEN l := SIZE( x ) END;
  IF b = NIL THEN Lib.Fill( ADR(x), l, 0 ); RETURN l END;
  IF l > b^.l THEN l := b^.l END;
  Enter( b^.m );
  Lib.Move( b^.a, ADR( x ), l );
  Leave( b^.m );
  Show();
  RETURN l;
END Read;


PROCEDURE ChkCfgs();

  PROCEDURE ChkName( i: TId );
  VAR
    n: TBaseName;
    j: TId;
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    n := cfg[i].n;
    FOR j := i TO MAX( TId ) DO
      IF ( i # j ) & ( Str.Compare( n, cfg[j].n ) = 0 ) THEN
        SO.Init( b, s );
        SO.WrStr( b, 'ChkName: Duplicate buffer name "' );
        SO.WrStr( b, n );
        SO.WrStr( b, '".' );
        Fatal( s );
      END;
    END;
  END ChkName;

  PROCEDURE ChkLen( i: TId );
  VAR
    l: CARDINAL;
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    l := cfg[i].l;
    IF l = 0 THEN  (* not allowed for CreateFileMapping *)
      SO.Init( b, s );
      SO.WrStr( b, 'ChkLen: Invalid length ' );
      SO.WrLngCard( b, l, 1 );
      SO.WrStr( b, ' for buffer "' );
      SO.WrStr( b, cfg[i].n );
      SO.WrStr( b, '".' );
      Fatal( s );
    END;
  END ChkLen;

VAR
  i: TId;
BEGIN
  FOR i := MIN( TId ) TO MAX( TId ) DO
    ChkName( i );
    ChkLen( i );
  END;
END ChkCfgs;

PROCEDURE InitModule();
BEGIN
  Log.Register( mid, 'ShdMem' );
  Lib2.Fill( bufs, 0 );
  ChkCfgs();
END InitModule;

BEGIN
  InitModule();
END ShdMem.
