(* shared memory *)
(* solyga@gmx.de, 2020-04-10 *)
IMPLEMENTATION MODULE ShdMem;

IMPORT WinDef, WinNT, WinBase;
IMPORT Lib, Str, Lib2;
IMPORT Types, SO, Cfg, Log;
FROM SYSTEM IMPORT BYTE, ADDRESS, ADR;
FROM Storage IMPORT ALLOCATE;

CONST
  mid = Cfg.midShdMem;

  pfx = 'Local\ShdMem_';  (* 'Global\' seems to require admin rights *)
  cntsName = 'Counts';

TYPE
  TResName = ARRAY [0..127] OF CHAR;

TYPE
  THandle = WinDef.HANDLE;

CONST
  hndInv = WinBase.INVALID_HANDLE_VALUE;  (* 0FFFFFFFFH *)

TYPE
  TMutex = THandle;

  TBufDesc = RECORD
    i: TId;
    l: CARDINAL;
    m: TMutex;
    a: ADDRESS;
  END;
  TBuf = POINTER TO TBufDesc;

  TWrtr = POINTER TO TWrtrDesc;
  TWrtrDesc = RECORD
    l: CARDINAL;
    m: TMutex;
    a: ADDRESS;
    b: TBuf;
  END;

  TRdr = POINTER TO TRdrDesc;
  TRdrDesc = RECORD
    l: CARDINAL;
    m: TMutex;
    a: ADDRESS;
    b: TBuf;
  END;

  TCntsDesc = RECORD
    m: TMutex;
    a: POINTER TO ARRAY TId OF RECORD w, r: CARDINAL END;  (* writer and reader counts *)
  END;

VAR
  bufs: ARRAY TId OF TBuf;
  cnts: TCntsDesc;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlMsg, s, FALSE );
END Message;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlFat, s, TRUE );
  HALT;
END Fatal;


PROCEDURE WrHex( VAR b: SO.TBuffer; x-: ARRAY OF BYTE; l: CARDINAL );
CONST
  lMax = 8;
VAR
  i: CARDINAL;
  c: CHAR;
BEGIN
  IF l > SIZE( x ) THEN l := SIZE( x ) END;
  IF l > lMax THEN l := lMax END;
  IF l = 0 THEN RETURN END;
  c := b.c.p;
  b.c.p := '0';
  DEC( l );
  FOR i := 0 TO l DO
    SO.WrShtHex( b, Types.TShtCard( x[i] ), 2 );
    IF i < l THEN SO.WrChar( b, ' ' ) END;
  END;
  b.c.p := c;
END WrHex;


PROCEDURE NewMtx( n-: ARRAY OF CHAR ): TMutex;
VAR
  h: THandle;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  h := WinBase.CreateMutex( NIL, FALSE, n );
  IF h = NIL THEN
    SO.Init( b, s );
    SO.WrStr( b, 'NewMtx: CreateMutex failed, error code ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  RETURN h;
END NewMtx;

PROCEDURE NewShm( n-: ARRAY OF CHAR; l: LONGCARD ): ADDRESS;
VAR
  h: THandle;
  a: ADDRESS;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  h := WinBase.CreateFileMapping(
    hndInv,                (* file handle *)
    NIL,                   (* attributes *)
    WinNT.PAGE_READWRITE,  (* protection *)
    0,                     (* size high *)
    l,                     (* size low, 0 (with high = 0) => WinError.ERROR_INVALID_PARAMETER = 87 *)
    n                      (* name *)
  );
  IF h = NIL THEN
    SO.Init( b, s );
    SO.WrStr( b, 'NewShm: CreateFileMapping failed, error code ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  (* WinError.ERROR_ALREADY_EXISTS = 183 *)
  SO.Init( b, s );
  SO.WrStr( b, 'NewShm: CreateFileMapping succeeded, error code ' );
  SO.WrLngCard( b, WinBase.GetLastError(), 1 );
  Message( s );
  (* docs.microsoft.com: "The initial contents of the pages in a file mapping object
       backed by the operating system paging file are 0 (zero)."
     SOL: It's impossible for the user to safely initialize the memory. OS help is required -
     either with interlocking or with a constant initialization. *)

  a := WinBase.MapViewOfFile(
    h,
    WinBase.FILE_MAP_WRITE,
    0,  (* offset high *)
    0,  (* offset low *)
    l   (* size (still unclear, since 1 works) *)
  );
  IF a = NIL THEN
    SO.Init( b, s );
    SO.WrStr( b, 'NewShm: MapViewOfFile failed, error code ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  SO.Init( b, s );
  SO.WrStr( b, 'NewShm: MapViewOfFile succeeded, adr ' );
  SO.WrLngHex( b, LONGCARD( a ), 8 );
  Message( s );

  RETURN a;
END NewShm;

PROCEDURE MkMtxName( VAR s: ARRAY OF CHAR; n-: ARRAY OF CHAR );
BEGIN
  Str.Copy( s, pfx );
  Str.Append( s, n );
  Str.Append( s, '_Mutex' );
END MkMtxName;

PROCEDURE MkShmName( VAR s: ARRAY OF CHAR; n-: ARRAY OF CHAR );
BEGIN
  Str.Copy( s, pfx );
  Str.Append( s, n );
  Str.Append( s, '_Buffer' );
END MkShmName;

PROCEDURE Buffer( i: TId ): TBuf;
VAR
  b: TBuf;
  l: CARDINAL;
  n: TResName;
BEGIN
  b := bufs[i];
  IF b # NIL THEN RETURN b END;
  NEW( b );
  Lib2.Fill( b^, 0 );
  b^.i := i;           l := cfg[i].l;
  b^.l := l;           MkMtxName( n, cfg[i].n );
  b^.m := NewMtx( n ); MkShmName( n, cfg[i].n );
  b^.a := NewShm( n, l );
  bufs[i] := b;
  RETURN b;
END Buffer;


(* the shared memory is not automatically interlocked by win (tested) *)
PROCEDURE Enter( m: TMutex );
VAR
  i: LONGCARD;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  i := WinBase.WaitForSingleObject( m, WinBase.INFINITE );
  IF i = MAX( LONGCARD ) THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Enter: WaitForSingleObject failed, err ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  IF i > 0 THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Enter: WaitForSingleObject returned ' );
    SO.WrLngCard( b, i, 1 );
    Fatal( s );
  END;
END Enter;

PROCEDURE Leave( m: TMutex );
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  IF ~ WinBase.ReleaseMutex( m ) THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Leave: ReleaseMutex failed, err ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
END Leave;


PROCEDURE IncCnt( i: TId; rdr: BOOLEAN );

  PROCEDURE Load();
  VAR
    n: TResName;
  BEGIN
    IF cnts.a # NIL THEN RETURN END;
    MkMtxName( n, cntsName );
    cnts.m := NewMtx( n );
    MkShmName( n, cntsName );
    cnts.a := NewShm( n, SIZE(cnts.a^) );
  END Load;

  PROCEDURE Err();
  TYPE
    TStrs = ARRAY BOOLEAN OF ARRAY [1..7] OF CHAR;
  CONST
    strs = TStrs{ 'writers', 'readers' };
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    SO.Init( b, s );
    SO.WrStr( b, 'IncCnt: Too many ' );
    SO.WrStr( b, strs[rdr] );
    SO.WrStr( b, ' for buffer "' );
    SO.WrStr( b, cfg[i].n );
    SO.WrStr( b, '".' );
    Fatal( s );
  END Err;

BEGIN
  Load();
  Enter( cnts.m );
  IF rdr THEN
    IF cnts.a^[i].r >= cfg[i].r THEN Err() END;
    INC( cnts.a^[i].r );
  ELSE
    IF cnts.a^[i].w >= cfg[i].w THEN Err() END;
    INC( cnts.a^[i].w );
  END;
  Leave( cnts.m );
END IncCnt;


PROCEDURE NewWrtr( i: TId ): TWrtr;
VAR
  b: TBuf;
  w: TWrtr;
BEGIN
  IncCnt( i, FALSE );
  b := Buffer( i );
  NEW( w );
  Lib2.Fill( w^, 0 );
  w^.l := b^.l;
  w^.m := b^.m;
  w^.a := b^.a;
  w^.b := b;
  RETURN w;
END NewWrtr;

PROCEDURE NewRdr( i: TId ): TRdr;
VAR
  b: TBuf;
  r: TRdr;
BEGIN
  IncCnt( i, TRUE );
  b := Buffer( i );
  NEW( r );
  Lib2.Fill( r^, 0 );
  r^.l := b^.l;
  r^.m := b^.m;
  r^.a := b^.a;
  r^.b := b;
  RETURN r;
END NewRdr;


PROCEDURE Write( w: TWrtr; x-: ARRAY OF BYTE; l: CARDINAL ): CARDINAL;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Write: ' );
    SO.WrLngCard( b, l, 1 );
    SO.WrStr( b, '  ' );
    WrHex( b, x, l );
    Message( s );
  END Show;

BEGIN
  IF l > SIZE( x ) THEN l := SIZE( x ) END;
  IF w = NIL THEN RETURN l END;
  IF l > w^.l THEN l := w^.l END;
  Enter( w^.m );
  Lib.Move( ADR( x ), w^.a, l );
  Leave( w^.m );
  Show();
  RETURN l;
END Write;

PROCEDURE Read( r: TRdr; VAR x: ARRAY OF BYTE; l: CARDINAL ): CARDINAL;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Read: ' );
    SO.WrLngCard( b, l, 1 );
    SO.WrStr( b, '  ' );
    WrHex( b, x, l );
    Message( s );
  END Show;

BEGIN
  IF l > SIZE( x ) THEN l := SIZE( x ) END;
  IF r = NIL THEN Lib.Fill( ADR(x), l, 0 ); RETURN l END;
  IF l > r^.l THEN l := r^.l END;
  Enter( r^.m );
  Lib.Move( r^.a, ADR( x ), l );
  Leave( r^.m );
  Show();
  RETURN l;
END Read;

PROCEDURE ChkName( i: TId );
VAR
  n: TBaseName;
  j: TId;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  n := cfg[i].n;
  IF Str.Compare( n, cntsName ) = 0 THEN
    SO.Init( b, s );
    SO.WrStr( b, 'ChkName: Invalid buffer name "' );
    SO.WrStr( b, n );
    SO.WrStr( b, '".' );
    Fatal( s );
  END;
  FOR j := i TO MAX( TId ) DO
    IF ( i # j ) & ( Str.Compare( n, cfg[j].n ) = 0 ) THEN
      SO.Init( b, s );
      SO.WrStr( b, 'ChkName: Duplicate buffer name "' );
      SO.WrStr( b, n );
      SO.WrStr( b, '".' );
      Fatal( s );
    END;
  END;
END ChkName;

PROCEDURE ChkLen( i: TId );
VAR
  l: CARDINAL;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  l := cfg[i].l;
  IF l = 0 THEN  (* not allowed for CreateFileMapping *)
    SO.Init( b, s );
    SO.WrStr( b, 'ChkLen: Invalid length ' );
    SO.WrLngCard( b, l, 1 );
    SO.WrStr( b, ' for buffer "' );
    SO.WrStr( b, cfg[i].n );
    SO.WrStr( b, '".' );
    Fatal( s );
  END;
END ChkLen;

PROCEDURE ChkCfgs();
VAR
  i: TId;
BEGIN
  FOR i := MIN( TId ) TO MAX( TId ) DO
    ChkName( i );
    ChkLen( i );
  END;
END ChkCfgs;

PROCEDURE InitModule();
BEGIN
  Log.Register( mid, 'ShdMem' );
  Lib2.Fill( bufs, 0 );
  Lib2.Fill( cnts, 0 );
  ChkCfgs();
END InitModule;

BEGIN
  InitModule();
END ShdMem.
